﻿//////////////////////////////////////////////
// BufferView.h
//
//////////////////////////////////////////////

/// Defines / Macros -------------------------

#pragma once

/// Includes ---------------------------------

// nkMemory
#include "../Dll/DllDefines.h"

#include "Buffer.h"
#include "BufferCast.h"

// Standards
#include <vector>

/// Class ------------------------------------

namespace nkMemory
{
	template <typename T = unsigned char>
	class BufferView
	{
		private :

			// Attributes
			T* _data ;
			unsigned long long _size ;

		public :

			// Constructor
			BufferView ()
			:	_data (nullptr),
				_size (0)
			{
				// Nothing to do
			}

			BufferView (T* data, unsigned long long size)
			:	_data (data),
				_size (size)
			{
				// Nothing to do
			}

			BufferView (const Buffer& buffer)
			:	_data ((T*)buffer.getData()),
				_size (buffer.getSize() / sizeof(T))
			{
				// Nothing to do
			}

			BufferView (const Buffer& buffer, unsigned long long index, unsigned long long length = 0)
			:	_data (&((T*)buffer.getData())[index]),
				_size (buffer.getSize() / sizeof(T) - index)
			{
				// Check length
				_size = length ? std::min(_size, length) : _size ;
			}

			// Block some translation that leads to trouble
			BufferView (Buffer&& buffer) = delete ;

			// Destructor
			// Doesn't free the data, merely a view
			~BufferView () = default ;

		public :

			// Getters
			T* getData () const
			{
				return _data ;
			}

			unsigned long long getSize () const
			{
				return _size ;
			}

			bool empty () const
			{
				return !_size ;
			}

			T& front ()
			{
				return _data[0] ;
			}

			const T& front () const
			{
				return _data[0] ;
			}

			T& back ()
			{
				return _data[_size - 1] ;
			}

			const T& back () const
			{
				return _data[_size - 1] ;
			}

		public :

			// Conversions
			BufferView<T> subView (unsigned long long index, unsigned long long length = 0) const
			{
				// Clamp size and length
				index = index < _size ? index : _size ;
				length = ((length > _size - index) || !length) ? _size - index : length ;

				if (!length)
					return BufferView<T>() ;

				return BufferView<T>(&_data[index], length) ;
			}

		public :

			// Operators
			T& operator[] (unsigned long long index)
			{
				return _data[index] ;
			}

			const T& operator[] (unsigned long long index) const
			{
				return _data[index] ;
			}

		public :

			// Constructors templated
			template <typename U>
			BufferView (const BufferCast<U>& bufferCast)
			:	_data ((T*)bufferCast.getData()),
				_size ((bufferCast.getSize() * sizeof(U)) / sizeof(T))
			{
				// Nothing to do
			}

			BufferView (const BufferCast<T>& bufferCast)
			:	_data (bufferCast.getData()),
				_size (bufferCast.getSize())
			{
				// Nothing to do
			}

			template <typename U>
			BufferView (const BufferView<U>& view)
			:	_data ((T*)view.getData()),
				_size ((view.getSize() * sizeof(U)) / sizeof(T))
			{
				// Nothing to do
			}

			BufferView (const BufferView<T>& view)
			:	_data (view._data),
				_size (view._size)
			{
				// Nothing to do
			}

			template <typename U>
			BufferView (std::vector<U>& vec)
			:	_data ((T*)vec.data()),
				_size ((vec.size() * sizeof(U)) / sizeof(T))
			{
				// Nothing to do
			}

			BufferView (std::vector<T>& vec)
			:	_data (vec.data()),
				_size (vec.size())
			{
				// Nothing to do
			}

			template <typename U, typename = std::enable_if<std::is_const<T>::value>>
			BufferView (const std::vector<U>& vec)
			:	_data (vec.data()),
				_size ((vec.size() * sizeof(U)) / sizeof(T))
			{
				// Nothing to do
			}

			template <typename = std::enable_if<std::is_const<T>::value>>
			BufferView (const std::vector<typename std::remove_const<T>::type>& vec)
			:	_data (vec.data()),
				_size (vec.size())
			{
				// Nothing to do
			}

		public :

			// Conversion operators
			operator Buffer () const
			{
				return nkMemory::Buffer((unsigned char*)_data, _size * sizeof(T)) ;
			}
	} ;
}

/// Exports ----------------------------------

namespace nkMemory
{
	// Unsigned
	template class DLL_MEMORY_EXPORT BufferView<unsigned char> ;
	template class DLL_MEMORY_EXPORT BufferView<unsigned short> ;
	template class DLL_MEMORY_EXPORT BufferView<unsigned int> ;
	template class DLL_MEMORY_EXPORT BufferView<unsigned long long> ;

	// Signed
	template class DLL_MEMORY_EXPORT BufferView<char> ;
	template class DLL_MEMORY_EXPORT BufferView<short> ;
	template class DLL_MEMORY_EXPORT BufferView<int> ;
	template class DLL_MEMORY_EXPORT BufferView<long long> ;

	// Floating
	template class DLL_MEMORY_EXPORT BufferView<float> ;
	template class DLL_MEMORY_EXPORT BufferView<double> ;
}